; 65c816 opcodes 0x40-0x4F:
;    40: RTI s       1,7
;    41: EOR (d,x)   2,6
;    42: WDM         2,2
;    43: EOR d,s     2,4
;    44: MVP xya     3,7
;    45: EOR d       2,3
;    46: LSR d       2,5
;    47: EOR [d]     2,6
;    48: PHA s       1,3
;    49: EOR #       2,2
;    4A: LSR A       1,2
;    4B: PHK s       1,3
;    4C: JMP a       3,3
;    4D: EOR a       3,4
;    4E: LSR a       3,6
;    4F: EOR al      4,5

; 40

ALIGNC
E0_RTI:
 E0_PULL_B
 E0_RESTOREFLAGS
 E0_PULL_W
 mov R_NativePC,eax
 E0_PULL_B
 mov B_PB,al
 test [MEMSEL],al   ; Check bus speed against PB
 mov bl,5
 jnz .fastrom
 mov bl,4
.fastrom:
 mov ah,0
 mov [CycleTable],bl    ; Update bus speed
 mov edi,R_NativePC
 shl eax,16
 or edi,eax

 shr edi,13
 mov edi,[_Read_Bank8Offset+edi*4]
 test edi,edi
 jnz .read_direct

 xor eax,eax
 mov edi,_Blank     ;we should really trap this...
.read_direct:
 add edi,eax

 add R_NativePC,edi
 mov B_PBOffset,edi

 JUMP_FLAG SNES_FLAG_I,.no_irq
 mov al,[IRQ_pin]
 test al,al
 jz .no_irq
 SAVE_PC ebx
 add R_Cycles,byte 60   ; IRQ processing: 2 IO + 6 bank0
 call E0_IRQ
 LOAD_PC
.no_irq:
 SET_TABLE_MX

ALIGNC
E1_RTI:
 E1_PULL_B
 E1_RESTOREFLAGS
 E1_PULL_W

 mov ebx,B_PB_Shifted
 mov R_NativePC,eax
 or eax,ebx

 shr eax,13
 mov eax,[_Read_Bank8Offset+eax*4]
 test eax,eax
 jnz .read_direct

 xor ebx,ebx
 mov eax,_Blank     ;we should really trap this...
.read_direct:
 add eax,ebx

 add R_NativePC,eax
 mov B_PBOffset,eax

 JUMP_FLAG SNES_FLAG_I,CPU_RETURN,near
 mov al,[IRQ_pin]
 test al,al
 jz near CPU_RETURN
 SAVE_PC ebx
 add R_Cycles,byte 52   ; IRQ processing: 2 IO + 5 bank0
 call E1_IRQ
 LOAD_PC
 jmp near CPU_RETURN

; 41
EM_EOR Od_xO,Direct_Indexed_Indirect
E0_EOR Od_xO,Direct_Indexed_Indirect

; 42
ALIGNC
ALL_WDM:
 add R_NativePC,byte 2
 jmp near CPU_RETURN

; 43
EM_EOR d_s,Stack_Relative
E0_EOR d_s,Stack_Relative

; 44

ALIGNC
E0_MVP:
 push ebx           ; Cycles per instruction iteration
.again:
 mov al,[1+R_NativePC]  ; Dest bank
 mov bl,[2+R_NativePC]  ; Src bank
 mov B_DB,al
 mov ax,B_X         ; Src address
 shl ebx,16
 add ebx,eax
 dec eax
 mov B_X,ax
 GET_BYTE
 mov ebx,B_DB_Shifted
 mov bx,B_Y        ; Dest address
 SET_BYTE
 mov eax,B_A
 dec ebx
 sub eax,byte 1
 mov B_Y,bx
 jc  .done
 cmp B_EventTrip,R_Cycles
 jbe .event
 mov B_A,ax
 add R_Cycles,[esp]
 xor eax,eax
 xor ebx,ebx
 jmp .again
ALIGNC
.done:
 add R_NativePC,byte 3
.event:
 mov B_A,ax
 pop edi
 jmp near CPU_RETURN
	
ALIGNC
EX_MVP:
 push ebx           ; Cycles per instruction iteration
.again:
 mov al,[1+R_NativePC]  ; Dest bank
 mov bl,[2+R_NativePC]  ; Src bank
 mov B_DB,al
 mov al,B_X         ; Src address
 shl ebx,16
 add ebx,eax
 dec eax
 mov B_X,al
 GET_BYTE
 mov ebx,B_DB_Shifted
 mov bl,B_Y         ; Dest address
 SET_BYTE
 mov eax,B_A
 dec ebx
 sub eax,byte 1
 mov B_Y,bl
 jc  .done
 cmp B_EventTrip,R_Cycles
 jbe .event
 mov B_A,ax
 add R_Cycles,[esp]
 xor eax,eax
 xor ebx,ebx
 jmp .again
ALIGNC
.done:
 add R_NativePC,byte 3
.event:
 mov B_A,ax
 pop edi
 jmp near CPU_RETURN

; 45
EM_EOR d,Direct
E0_EOR d,Direct

; 46
EM_LSR d,Direct
E0_LSR d,Direct

; 47
EM_EOR IdI,Direct_Indirect_Long
E0_EOR IdI,Direct_Indirect_Long

; 48

ALIGNC
E0_PHA:
 inc R_NativePC
 mov eax,B_A
 E0_PUSH_W
 jmp near CPU_RETURN

ALIGNC
EM_PHA:
 inc R_NativePC
 mov eax,B_A
 E0_PUSH_B
 jmp near CPU_RETURN

ALIGNC
E1_PHA:
 inc R_NativePC
 mov eax,B_A
 E1_PUSH_B
 jmp near CPU_RETURN

; 49
EM_EOR i,Immediate
E0_EOR i,Immediate

; 4A

ALIGNC
EM_SRA:
 inc R_NativePC
 shr byte B_A,byte 1
 STORE_FLAGS SNES_FLAG_NZC
 jmp near CPU_RETURN
ALIGNC
E0_SRA:
 inc R_NativePC
 shr word B_A,byte 1
 STORE_FLAGS SNES_FLAG_NZC
 jmp near CPU_RETURN

; 4B

ALIGNC
E0_PHK:             ; Push PB(K) onto stack
 inc R_NativePC
 mov al,B_PB
 E0_PUSH_B
 jmp near CPU_RETURN

ALIGNC
E1_PHK:
 inc R_NativePC
 mov al,B_PB
 E1_PUSH_B
 jmp near CPU_RETURN

; 4C

ALIGNC
ALL_JMP_a:
 ADDR_Absolute_JMP

 mov ebx,B_PB_Shifted
 mov R_NativePC,eax
 or eax,ebx

 shr eax,13
 mov eax,[_Read_Bank8Offset+eax*4]
 test eax,eax
 jnz .read_direct

 xor ebx,ebx
 mov eax,_Blank     ;we should really trap this...
.read_direct:
 add eax,ebx

 add R_NativePC,eax
 mov B_PBOffset,eax

 jmp near CPU_RETURN

; 4D
EM_EOR a,Absolute
E0_EOR a,Absolute

; 4E
EM_LSR a,Absolute
E0_LSR a,Absolute

; 4F
EM_EOR al,Absolute_Long
E0_EOR al,Absolute_Long
