; SNEeSe 65c816 CPU emulation core
; Originally written by Savoury SnaX (Not quite sure if I like AT&T)
; Maintained/rewritten by Charles Bilyue'
;
; Compile under DJGPP (GNU-AS)
;  NOTE - This must have .S (capital) or the # stuff wont be pre-processed!!!
;
; This file contains:
;  CPU opcode macros
;
; Examples of usage:
;  EM_LDA Direct = 8-bit LDA using "direct" address mode
;  BFC SNES_FLAG_C = Branch if carry flag clear (BCC)

;%1 = label, %2 = addr
%macro EM_ADC 2
ALIGNC
EM_ADC_%1:
 READ8_%2
 CLR_FLAG SNES_FLAG_V
 JUMP_FLAG SNES_FLAG_D,%%decimal_mode
 MAKE_CARRY
 adc B_A,al
 lahf
 jno %%no_overflow
 SET_FLAG SNES_FLAG_V
%%no_overflow:
 STORE_FLAGS_FROM SNES_FLAG_NZC,ah
 jmp near CPU_RETURN

ALIGNC
%%decimal_mode:
 MAKE_CARRY
 adc al,B_A
 daa
 mov B_A,al
 lahf
 jno %%no_overflow_decimal
%%overflow:
 SET_FLAG SNES_FLAG_V
%%no_overflow_decimal:
 STORE_FLAGS_FROM SNES_FLAG_NZC,ah
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_ADC 2
ALIGNC
E0_ADC_%1:
 READ16_%2
 CLR_FLAG SNES_FLAG_V
 JUMP_FLAG SNES_FLAG_D,%%decimal_mode
 MAKE_CARRY
 adc B_A,ax
 lahf
 jno %%no_overflow
 SET_FLAG SNES_FLAG_V
%%no_overflow:
 STORE_FLAGS_FROM SNES_FLAG_NZC,ah
 jmp near CPU_RETURN

ALIGNC
%%decimal_mode:
 MAKE_CARRY
 mov ebx,eax
 mov eax,B_A
 adc al,bl
 daa
 mov bl,al
 mov al,ah
 adc al,bh
 daa
 mov bh,al
 mov B_A,bx
 lahf
 jno %%no_overflow_decimal
 SET_FLAG SNES_FLAG_V
%%no_overflow_decimal:
 STORE_FLAGS_FROM SNES_FLAG_C,ah
;test bx,bx
 shl ebx,16
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_AND 2
ALIGNC
EM_AND_%1:
 READ8_%2
 and B_A,al
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_AND 2
ALIGNC
E0_AND_%1:
 READ16_%2
 and B_A,ax
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_ASL 2
ALIGNC
EM_ASL_%1:
 READ8_%2
 add al,al
 STORE_FLAGS SNES_FLAG_NZC
 SET_BYTE
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_ASL 2
ALIGNC
E0_ASL_%1:
 READ16_%2
 mov edi,eax
 add di,di
 STORE_FLAGS SNES_FLAG_NZC
 mov eax,edi
 SET_WORD_HL
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = flag
%macro BFC 2
ALIGNC
ALL_%1:
 add R_NativePC,byte 2
 JUMP_FLAG %2,CPU_RETURN,near       ; flag set
 movsx eax,byte [R_NativePC - 1]    ; sign extend for addition
 add R_Cycles,byte 6            ; branch taken
 add R_NativePC,eax
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = flag
%macro BFS 2
ALIGNC
ALL_%1:
 add R_NativePC,byte 2
 JUMP_NOT_FLAG %2,CPU_RETURN,near   ; flag clear
 movsx eax,byte [R_NativePC - 1]    ; sign extend for addition
 add R_Cycles,byte 6            ; branch taken
 add R_NativePC,eax
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_BIT 2
ALIGNC
EM_BIT_%1:
 READ8_%2
 CLR_FLAG (SNES_FLAG_N+SNES_FLAG_V+SNES_FLAG_Z)
 test al,al
 jns %%no_negative
 SET_FLAG SNES_FLAG_N
%%no_negative:
 test al,0x40
 jz  %%no_overflow
 SET_FLAG SNES_FLAG_V
%%no_overflow:
 test al,B_A
 jnz %%no_zero
 SET_FLAG SNES_FLAG_Z
%%no_zero:
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_BIT 2
ALIGNC
E0_BIT_%1:
 READ16_%2
 CLR_FLAG (SNES_FLAG_N+SNES_FLAG_V+SNES_FLAG_Z)
 test ah,ah
 jns %%no_negative
 SET_FLAG SNES_FLAG_N
%%no_negative:
 test ah,0x40
 jz %%no_overflow
 SET_FLAG SNES_FLAG_V
%%no_overflow:
 test eax,B_A
 jnz %%no_zero
 SET_FLAG SNES_FLAG_Z
%%no_zero:
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_CMP 2
ALIGNC
EM_CMP_%1:
 READ8_%2
 cmp B_A,al
 cmc
 STORE_FLAGS SNES_FLAG_NZC
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_CMP 2
ALIGNC
E0_CMP_%1:
 READ16_%2
 cmp B_A,ax
 cmc
 STORE_FLAGS SNES_FLAG_NZC
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EX_CPX 2
ALIGNC
EX_CPX_%1:
 READ8_%2
 cmp B_X,al
 cmc
 STORE_FLAGS SNES_FLAG_NZC
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_CPX 2
ALIGNC
E0_CPX_%1:
 READ16_%2
 cmp B_X,ax
 cmc
 STORE_FLAGS SNES_FLAG_NZC
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EX_CPY 2
ALIGNC
EX_CPY_%1:
 READ8_%2
 cmp B_Y,al
 cmc
 STORE_FLAGS SNES_FLAG_NZC
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_CPY 2
ALIGNC
E0_CPY_%1:
 READ16_%2
 cmp B_Y,ax
 cmc
 STORE_FLAGS SNES_FLAG_NZC
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_DEC 2
ALIGNC
EM_DEC_%1:
 READ8_%2
 dec eax
 SET_BYTE
 test al,al
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_DEC 2
ALIGNC
E0_DEC_%1:
 READ16_%2
 dec eax
 SET_WORD_HL
;test ax,ax
 shl eax,16
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_EOR 2
ALIGNC
EM_EOR_%1:
 READ8_%2
 xor B_A,al
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_EOR 2
ALIGNC
E0_EOR_%1:
 READ16_%2
 xor B_A,ax
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_INC 2
ALIGNC
EM_INC_%1:
 READ8_%2
 inc eax
 SET_BYTE
 test al,al
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_INC 2
ALIGNC
E0_INC_%1:
 READ16_%2
 inc eax
 SET_WORD_HL
;test ax,ax
 shl eax,16
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_LDA 2
ALIGNC
EM_LDA_%1:
 READ8_%2
 mov B_A,al
 test al,al
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_LDA 2
ALIGNC
E0_LDA_%1:
 READ16_%2
 mov B_A,ax
;test ax,ax
 shl eax,16
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EX_LDX 2
ALIGNC
EX_LDX_%1:
 READ8_%2
 mov B_X,al
 test al,al
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_LDX 2
ALIGNC
E0_LDX_%1:
 READ16_%2
 mov B_X,ax
;test ax,ax
 shl eax,16
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EX_LDY 2
ALIGNC
EX_LDY_%1:
 READ8_%2
 mov B_Y,al
 test al,al
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_LDY 2
ALIGNC
E0_LDY_%1:
 READ16_%2
 mov B_Y,ax
;test ax,ax
 shl eax,16
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_LSR 2
ALIGNC
EM_LSR_%1:
 READ8_%2
 shr al,byte 1
 STORE_FLAGS SNES_FLAG_NZC
 SET_BYTE
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_LSR 2
ALIGNC
E0_LSR_%1:
 READ16_%2
 mov edi,eax
 shr di,byte 1
 STORE_FLAGS SNES_FLAG_NZC
 mov eax,edi
 SET_WORD_HL
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_ORA 2
ALIGNC
EM_ORA_%1:
 READ8_%2
 or B_A,al
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_ORA 2
ALIGNC
E0_ORA_%1:
 READ16_%2
 or B_A,ax
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_ROL 2
ALIGNC
EM_ROL_%1:
 READ8_%2
 MAKE_CARRY
 adc al,al
 STORE_FLAGS SNES_FLAG_NZC
 SET_BYTE
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_ROL 2
ALIGNC
E0_ROL_%1:
 READ16_%2
 mov edi,eax
 MAKE_CARRY
 adc di,di
 STORE_FLAGS SNES_FLAG_NZC
 mov eax,edi
 SET_WORD_HL
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_ROR 2
ALIGNC
EM_ROR_%1:
 READ8_%2
 shr R_P_B,byte 1    ;MAKE_CARRY
 rcr al,byte 1
 adc R_P_B,R_P_B     ;STORE_FLAGS SNES_FLAG_C
 test al,al
 STORE_FLAGS SNES_FLAG_NZ
 SET_BYTE
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_ROR 2
ALIGNC
E0_ROR_%1:
 READ16_%2
 mov edi,eax
 shr R_P_B,byte 1   ;MAKE_CARRY
 rcr di,byte 1
 adc R_P_B,R_P_B    ;STORE_FLAGS SNES_FLAG_C
 test di,di
 STORE_FLAGS SNES_FLAG_NZ
 mov eax,edi
 SET_WORD_HL
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_SBC 2
ALIGNC
EM_SBC_%1:
 READ8_%2
 CLR_FLAG SNES_FLAG_V
 JUMP_FLAG SNES_FLAG_D,%%decimal_mode
 MAKE_NOT_CARRY
 sbb B_A,al
 cmc
 lahf
 jno %%no_overflow
 SET_FLAG SNES_FLAG_V
%%no_overflow:
 STORE_FLAGS_FROM SNES_FLAG_NZC,ah
 jmp near CPU_RETURN

ALIGNC
%%decimal_mode:
 MAKE_NOT_CARRY
 mov ah,B_A
 sbb ah,al
 mov al,ah
 das
 mov B_A,al
 cmc
 lahf
 jno %%no_overflow_decimal
%%overflow:
 SET_FLAG SNES_FLAG_V
%%no_overflow_decimal:
 STORE_FLAGS_FROM SNES_FLAG_NZC,ah
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_SBC 2
ALIGNC
E0_SBC_%1:
 READ16_%2
 CLR_FLAG SNES_FLAG_V
 JUMP_FLAG SNES_FLAG_D,%%decimal_mode
 MAKE_NOT_CARRY
 sbb B_A,ax
 cmc
 lahf
 jno %%no_overflow
 SET_FLAG SNES_FLAG_V
%%no_overflow:
 STORE_FLAGS_FROM SNES_FLAG_NZC,ah
 jmp near CPU_RETURN

ALIGNC
%%decimal_mode:
 MAKE_NOT_CARRY
 mov ebx,eax
 mov eax,B_A
 sbb al,bl
 das
 mov bl,al
 mov al,ah
 sbb al,bh
 das
 mov bh,al
 mov B_A,bx
 cmc
 lahf
 jno %%no_overflow_decimal
 SET_FLAG SNES_FLAG_V
%%no_overflow_decimal:
 STORE_FLAGS_FROM SNES_FLAG_C,ah
;test bx,bx
 shl ebx,16
 STORE_FLAGS SNES_FLAG_NZ
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_STA 2
ALIGNC
EM_STA_%1:
 ADDR_%2
 mov eax,B_A
 SET_BYTE
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_STA 2
ALIGNC
E0_STA_%1:
 ADDR_%2
 mov eax,B_A
 SET_WORD NoSave
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EX_STX 2
ALIGNC
EX_STX_%1:
 ADDR_%2
 mov eax,B_X
 SET_BYTE
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_STX 2
ALIGNC
E0_STX_%1:
 ADDR_%2
 mov eax,B_X
 SET_WORD NoSave
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EX_STY 2
ALIGNC
EX_STY_%1:
 ADDR_%2
 mov eax,B_Y
 SET_BYTE
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_STY 2
ALIGNC
E0_STY_%1:
 ADDR_%2
 mov eax,B_Y
 SET_WORD NoSave
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro EM_STZ 2
ALIGNC
EM_STZ_%1:
 ADDR_%2
 xor eax,eax
 SET_BYTE
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_STZ 2
ALIGNC
E0_STZ_%1:
 ADDR_%2
 xor eax,eax
 SET_WORD Double
 jmp near CPU_RETURN
%endmacro

; TRB set flag Z on A & Memory Operand not A & !Memory Operand!
;%1 = label, %2 = addr
%macro EM_TRB 2
ALIGNC
EM_TRB_%1:
 READ8_%2
 test B_A,al
 STORE_FLAGS SNES_FLAG_Z
 mov ah,B_A
 xor ah,-1
 and al,ah
 SET_BYTE
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_TRB 2
ALIGNC
E0_TRB_%1:
 READ16_%2
 mov edi,eax
 test edi,B_A
 STORE_FLAGS SNES_FLAG_Z
 mov eax,B_A
 xor eax,byte -1
 and eax,edi
 SET_WORD_HL
 jmp near CPU_RETURN
%endmacro

; TSB set flag Z on A & Memory Operand not A | Memory Operand!

;%1 = label, %2 = addr
%macro EM_TSB 2
ALIGNC
EM_TSB_%1:
 READ8_%2
 test B_A,al
 STORE_FLAGS SNES_FLAG_Z
 or al,B_A
 SET_BYTE
 jmp near CPU_RETURN
%endmacro

;%1 = label, %2 = addr
%macro E0_TSB 2
ALIGNC
E0_TSB_%1:
 READ16_%2
 mov edi,eax
 test edi,B_A
 STORE_FLAGS SNES_FLAG_Z
 mov eax,B_A
 or eax,edi
 SET_WORD_HL
 jmp near CPU_RETURN
%endmacro
