#include <string.h>
#include "wrapaleg.h"
#include "helper.h"

#define OUTPUT_AUDIO_UNSIGNED

#ifdef OUTPUT_AUDIO_UNSIGNED
#define OUTPUT_ZERO_BASE 0x8000
#define OUTPUT_LOWER_LIMIT 0
#define OUTPUT_UPPER_LIMIT 0xFFFF
typedef unsigned short output_sample;
#else
#define OUTPUT_ZERO_BASE 0
#define OUTPUT_LOWER_LIMIT 0x8000
#define OUTPUT_UPPER_LIMIT 0x7FFF
typedef short output_sample;
#endif

#define BRR_PACKET_END  (1 << 0)
#define BRR_PACKET_LOOP (1 << 1)

#define FREQ 32000      //32kHz true SNES mix speed
#define LAG  10         //Lag between sound mixed and sound heard (1/LAG sec)

#define ADSR
//#define ADSR_OPTIONAL

#define ATTACK 0    // A of ADSR
#define DECAY 1     // D of ADSR
#define SUSTAIN 2   // S of ADSR
#define RELEASE 3   // R of ADSR
#define DECREASE 4  // GAIN linear decrease mode
#define EXP 5       // GAIN exponential decrease mode
#define INCREASE 6  // GAIN linear increase mode
#define BENT 7      // GAIN bent line increase mode
#define DIRECT 8    // Directly specify ENVX

#define DSP_MAIN_LVOL   0x0C
#define DSP_MAIN_RVOL   0x1C
#define DSP_ECHO_LVOL   0x2C
#define DSP_ECHO_RVOL   0x3C
#define DSP_KON         0x4C
#define DSP_KOF         0x5C
#define DSP_FLG         0x6C
#define DSP_ENDX        0x7C

#define DSP_EFB     0x0D
#define DSP_PMON    0x2D
#define DSP_NOV     0x3D
#define DSP_EOV     0x4D
#define DSP_DIR     0x5D
#define DSP_ESA     0x6D
#define DSP_EDL     0x7D

#define DSP_VOICE_LVOL      0
#define DSP_VOICE_RVOL      1
#define DSP_VOICE_PITCH_L   2
#define DSP_VOICE_PITCH_H   3
#define DSP_VOICE_SRCN      4
#define DSP_VOICE_ADSR1     5
#define DSP_VOICE_ADSR2     6
#define DSP_VOICE_GAIN      7
#define DSP_VOICE_ENVX      8
#define DSP_VOICE_OUTX      9

#define ENVX_MAX 0x7FFFFF00
#define SPC_CLOCK_HZ 2048000

extern unsigned char SPCRAM[65536];
extern unsigned long SPC_Cycles;

unsigned TotalCycles;

unsigned short SPC_T0_position, SPC_T1_position, SPC_T2_position;
unsigned short SPC_T0_target, SPC_T1_target, SPC_T2_target;
unsigned SPC_T0_cycle_latch, SPC_T1_cycle_latch, SPC_T2_cycle_latch;
unsigned char SPC_T0_counter, SPC_T1_counter, SPC_T2_counter;

unsigned char SPC_CTRL, SPC_MASK;
unsigned char SPC_DSP_ADDR, SPC_DSP_DATA;

unsigned sound_cycle_latch;
unsigned sound_output_position;

int voice_handle;
signed char block_written[2];

static SAMPLE *STRsamp, *TMPsamp, *NOISEsamp;

#ifdef ADSR_OPTIONAL
signed char ADSR_ENABLED = -1;
#endif
signed char SOUND_ENABLED;
signed char ENVX_ENABLED = -1;

unsigned char SPC_DSP[128];

// INTERNAL (static) STUFF

static struct voice_state {
    short buf[16];
    unsigned envx, envcyc, sound_cycle_latch, env_update_time;
    unsigned ar, dr, sl, sr;
    int envstate;
    int gn, gain_update_time;
    unsigned long bufptr;
    unsigned brrptr;
    unsigned short *loopptr;
    short last1, last2;
} SNDvoices[8];

static int SNDkeys;

/* How many cycles till ADSR/GAIN adjustment
   Educated guesses on most of these
 */

static const unsigned attack_time[16] =
{
 (unsigned) (SPC_CLOCK_HZ * 4.100 / 64), (unsigned) (SPC_CLOCK_HZ * 2.600 / 64),
 (unsigned) (SPC_CLOCK_HZ * 1.500 / 64), (unsigned) (SPC_CLOCK_HZ * 1.000 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.640 / 64), (unsigned) (SPC_CLOCK_HZ * 0.380 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.260 / 64), (unsigned) (SPC_CLOCK_HZ * 0.160 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.096 / 64), (unsigned) (SPC_CLOCK_HZ * 0.064 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.040 / 64), (unsigned) (SPC_CLOCK_HZ * 0.024 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.016 / 64), (unsigned) (SPC_CLOCK_HZ * 0.010 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.006 / 64), 0
};

static const unsigned decay_time[16] =
{
 (unsigned) (SPC_CLOCK_HZ * 1.200 / 619), (unsigned) (SPC_CLOCK_HZ * 0.740 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.440 / 619), (unsigned) (SPC_CLOCK_HZ * 0.290 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.180 / 619), (unsigned) (SPC_CLOCK_HZ * 0.110 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.074 / 619), (unsigned) (SPC_CLOCK_HZ * 0.037 / 619)
};

static const unsigned sustain_time[32] =
{
 0 - 1                                  , (unsigned) (SPC_CLOCK_HZ * 38.00 / 619),
 (unsigned) (SPC_CLOCK_HZ * 28.00 / 619), (unsigned) (SPC_CLOCK_HZ * 24.00 / 619),
 (unsigned) (SPC_CLOCK_HZ * 19.00 / 619), (unsigned) (SPC_CLOCK_HZ * 14.00 / 619),
 (unsigned) (SPC_CLOCK_HZ * 12.00 / 619), (unsigned) (SPC_CLOCK_HZ * 9.400 / 619),
 (unsigned) (SPC_CLOCK_HZ * 7.100 / 619), (unsigned) (SPC_CLOCK_HZ * 5.900 / 619),
 (unsigned) (SPC_CLOCK_HZ * 4.700 / 619), (unsigned) (SPC_CLOCK_HZ * 3.500 / 619),
 (unsigned) (SPC_CLOCK_HZ * 2.900 / 619), (unsigned) (SPC_CLOCK_HZ * 2.400 / 619),
 (unsigned) (SPC_CLOCK_HZ * 1.800 / 619), (unsigned) (SPC_CLOCK_HZ * 1.500 / 619),
 (unsigned) (SPC_CLOCK_HZ * 1.200 / 619), (unsigned) (SPC_CLOCK_HZ * 0.880 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.740 / 619), (unsigned) (SPC_CLOCK_HZ * 0.590 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.440 / 619), (unsigned) (SPC_CLOCK_HZ * 0.370 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.290 / 619), (unsigned) (SPC_CLOCK_HZ * 0.220 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.180 / 619), (unsigned) (SPC_CLOCK_HZ * 0.150 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.110 / 619), (unsigned) (SPC_CLOCK_HZ * 0.092 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.074 / 619), (unsigned) (SPC_CLOCK_HZ * 0.055 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.037 / 619), (unsigned) (SPC_CLOCK_HZ * 0.028 / 619)
};

static const unsigned linear_time[32] =
{
 0 - 1                                 , (unsigned) (SPC_CLOCK_HZ * 4.100 / 64),
 (unsigned) (SPC_CLOCK_HZ * 3.100 / 64), (unsigned) (SPC_CLOCK_HZ * 2.600 / 64),
 (unsigned) (SPC_CLOCK_HZ * 2.000 / 64), (unsigned) (SPC_CLOCK_HZ * 1.500 / 64),
 (unsigned) (SPC_CLOCK_HZ * 1.300 / 64), (unsigned) (SPC_CLOCK_HZ * 1.000 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.770 / 64), (unsigned) (SPC_CLOCK_HZ * 0.640 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.510 / 64), (unsigned) (SPC_CLOCK_HZ * 0.380 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.320 / 64), (unsigned) (SPC_CLOCK_HZ * 0.260 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.190 / 64), (unsigned) (SPC_CLOCK_HZ * 0.160 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.130 / 64), (unsigned) (SPC_CLOCK_HZ * 0.096 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.080 / 64), (unsigned) (SPC_CLOCK_HZ * 0.064 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.048 / 64), (unsigned) (SPC_CLOCK_HZ * 0.040 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.032 / 64), (unsigned) (SPC_CLOCK_HZ * 0.024 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.020 / 64), (unsigned) (SPC_CLOCK_HZ * 0.016 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.012 / 64), (unsigned) (SPC_CLOCK_HZ * 0.010 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.008 / 64), (unsigned) (SPC_CLOCK_HZ * 0.006 / 64),
 (unsigned) (SPC_CLOCK_HZ * 0.004 / 64), (unsigned) (SPC_CLOCK_HZ * 0.002 / 64)
};

static const unsigned exp_time[32] =
{
 0 - 1                                  , (unsigned) (SPC_CLOCK_HZ * 38.00 / 619),
 (unsigned) (SPC_CLOCK_HZ * 28.00 / 619), (unsigned) (SPC_CLOCK_HZ * 24.00 / 619),
 (unsigned) (SPC_CLOCK_HZ * 19.00 / 619), (unsigned) (SPC_CLOCK_HZ * 14.00 / 619),
 (unsigned) (SPC_CLOCK_HZ * 12.00 / 619), (unsigned) (SPC_CLOCK_HZ * 9.400 / 619),
 (unsigned) (SPC_CLOCK_HZ * 7.100 / 619), (unsigned) (SPC_CLOCK_HZ * 5.900 / 619),
 (unsigned) (SPC_CLOCK_HZ * 4.700 / 619), (unsigned) (SPC_CLOCK_HZ * 3.500 / 619),
 (unsigned) (SPC_CLOCK_HZ * 2.900 / 619), (unsigned) (SPC_CLOCK_HZ * 2.400 / 619),
 (unsigned) (SPC_CLOCK_HZ * 1.800 / 619), (unsigned) (SPC_CLOCK_HZ * 1.500 / 619),
 (unsigned) (SPC_CLOCK_HZ * 1.200 / 619), (unsigned) (SPC_CLOCK_HZ * 0.880 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.740 / 619), (unsigned) (SPC_CLOCK_HZ * 0.590 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.440 / 619), (unsigned) (SPC_CLOCK_HZ * 0.370 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.290 / 619), (unsigned) (SPC_CLOCK_HZ * 0.220 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.180 / 619), (unsigned) (SPC_CLOCK_HZ * 0.150 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.110 / 619), (unsigned) (SPC_CLOCK_HZ * 0.092 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.074 / 619), (unsigned) (SPC_CLOCK_HZ * 0.055 / 619),
 (unsigned) (SPC_CLOCK_HZ * 0.037 / 619), (unsigned) (SPC_CLOCK_HZ * 0.018 / 619)
};

static const unsigned bent_time[32] =
{
 0 - 1                                 , (unsigned) (SPC_CLOCK_HZ * 7.200 / 55),
 (unsigned) (SPC_CLOCK_HZ * 5.400 / 55), (unsigned) (SPC_CLOCK_HZ * 4.600 / 55),
 (unsigned) (SPC_CLOCK_HZ * 3.500 / 55), (unsigned) (SPC_CLOCK_HZ * 2.600 / 55),
 (unsigned) (SPC_CLOCK_HZ * 2.300 / 55), (unsigned) (SPC_CLOCK_HZ * 1.800 / 55),
 (unsigned) (SPC_CLOCK_HZ * 1.300 / 55), (unsigned) (SPC_CLOCK_HZ * 1.100 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.900 / 55), (unsigned) (SPC_CLOCK_HZ * 0.670 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.580 / 55), (unsigned) (SPC_CLOCK_HZ * 0.450 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.340 / 55), (unsigned) (SPC_CLOCK_HZ * 0.280 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.220 / 55), (unsigned) (SPC_CLOCK_HZ * 0.170 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.140 / 55), (unsigned) (SPC_CLOCK_HZ * 0.110 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.084 / 55), (unsigned) (SPC_CLOCK_HZ * 0.070 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.056 / 55), (unsigned) (SPC_CLOCK_HZ * 0.042 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.035 / 55), (unsigned) (SPC_CLOCK_HZ * 0.028 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.021 / 55), (unsigned) (SPC_CLOCK_HZ * 0.018 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.014 / 55), (unsigned) (SPC_CLOCK_HZ * 0.011 / 55),
 (unsigned) (SPC_CLOCK_HZ * 0.007 / 55), (unsigned) (SPC_CLOCK_HZ * 0.0035 / 55)
};

static const unsigned noise_frequencies[32] =
{
 0,16,21,25,31,42,50,63,84,100,125,167,200,250,333,400,500,667,800,
 1000,1300,1600,2000,2700,3200,4000,5300,6400,8000,10700,16000,32000
};
static const unsigned noise_freq_factors[32] =
{
 0,2048,1536,1280,1024,768,640,512,384,320,256,192,160,
 128,96,80,64,48,40,32,24,20,16,12,10,8,6,5,4,3,2,1
};

#define NOISE_FEEDBACK 0x00040001
int noise;
int noise_vol;
unsigned noise_freq_factor, noise_freq_latch;

// OLD STUFF

/* Timers 0/1 must be updated once every 10M, 2 every 1.28M SPC cycles */
void Update_SPC_Timer_0()
{
 if (!(SPC_CTRL & 1)) return;
 SPC_T0_position += (TotalCycles - SPC_T0_cycle_latch) / 256;
 SPC_T0_cycle_latch += (TotalCycles - SPC_T0_cycle_latch) & ~0xFF;

 if (SPC_T0_position < SPC_T0_target) return;
 SPC_T0_counter = (SPC_T0_counter + (SPC_T0_position / SPC_T0_target)) & 0xF;
 SPC_T0_position %= SPC_T0_target;
}

void Update_SPC_Timer_1()
{
 if (!(SPC_CTRL & 2)) return;
 SPC_T1_position += (TotalCycles - SPC_T1_cycle_latch) / 256;
 SPC_T1_cycle_latch += (TotalCycles - SPC_T1_cycle_latch) & ~0xFF;

 if (SPC_T1_position < SPC_T1_target) return;
 SPC_T1_counter = (SPC_T1_counter + (SPC_T1_position / SPC_T1_target)) & 0xF;
 SPC_T1_position %= SPC_T1_target;
}

void Update_SPC_Timer_2()
{
 if (!(SPC_CTRL & 4)) return;
 SPC_T2_position += (TotalCycles - SPC_T2_cycle_latch) / 32;
 SPC_T2_cycle_latch += (TotalCycles - SPC_T2_cycle_latch) & ~0x1F;

 if (SPC_T2_position < SPC_T2_target) return;
 SPC_T2_counter = (SPC_T2_counter + (SPC_T2_position / SPC_T2_target)) & 0xF;
 SPC_T2_position %= SPC_T2_target;
}

void Wrap_SPC_Cyclecounter(){
 int c;

 for (c = 0; c < 8; c++)
 {
  SNDvoices[c].envcyc -= 0xF0000000;
  SNDvoices[c].sound_cycle_latch -= 0xF0000000;
 }
 TotalCycles -= 0xF0000000;
 SPC_Cycles -= 0xF0000000;
 SPC_T0_cycle_latch -= 0xF0000000;
 SPC_T1_cycle_latch -= 0xF0000000;
 SPC_T2_cycle_latch -= 0xF0000000;
 sound_cycle_latch -= 0xF0000000;
}

// NEW PRIVATE (static) STUFF

static int get_chunk(int voice)
{
    struct voice_state *pvs;
    int i, output;
    int last1, last2;
    unsigned int brrptr;
    unsigned char header,range,filter,input;

    pvs = &SNDvoices[voice];

    brrptr = pvs->brrptr;
    last1 = pvs->last1;
    last2 = pvs->last2;

    if (brrptr >= 0x10000 - 9)
    {
     SPC_DSP[DSP_ENDX] |= 1 << voice;
     pvs->brrptr = brrptr;
     pvs->last1 = last1;
     pvs->last2 = last2;
     return 1;
    }

    header = SPCRAM[brrptr++];
    filter = (header >> 2) & 3;
    range = header >> 4;
    i = 0;

    while (i < 16)
    {

        input = SPCRAM[brrptr++];

        output = (input >> 4) & 0x0F;
        output = (output ^ 8) - 8;
        output <<= range;

        switch (filter)
        {
        case 1:
            output += last1 * 15 / 16;
            break;
        case 2:
            output += (last1 * 61 / 32) - (last2 * 15 / 16);
            break;
        case 3:
            output += (last1 * 115 / 64) - (last2 * 13 / 16);
            break;
        }

        // Clip underflow/overflow (saturation)
        if (output > 0x7FFF)
            output = 0x7FFF;
        else if (output < -0x8000)
            output = -0x8000;

        last2 = last1;
        last1 = output;
        pvs->buf[i++] = output;

        output = input & 0x0F;
        output = (output ^ 8) - 8;
        output <<= range;

        switch (filter)
        {
        case 1:
            output += last1 * 15 / 16;
            break;
        case 2:
            output += (last1 * 61 / 32) - (last2 * 15 / 16);
            break;
        case 3:
            output += (last1 * 115 / 64) - (last2 * 13 / 16);
            break;
        }

        // Clip underflow/overflow (saturation)
        if (output > 0x7FFF)
            output = 0x7FFF;
        else if (output < -0x8000)
            output = -0x8000;

        last2 = last1;
        last1 = output;
        pvs->buf[i++] = output;
    }

    if (header & BRR_PACKET_END)
    {
        if (header & BRR_PACKET_LOOP)
            pvs->brrptr = *pvs->loopptr;
        else
            pvs->brrptr = 0x10000;

        SPC_DSP[DSP_ENDX] |= 1 << voice;
    }
    else
    {
        pvs->brrptr = brrptr;
    }

    pvs->last1 = last1;
    pvs->last2 = last2;
    return 0;
}

static int SNDDoEnv(int voice)
{
    struct voice_state *pvs;
    unsigned envx;
    unsigned voice_cycle_latch;

    pvs = &SNDvoices[voice];
    envx = pvs->envx;
    voice_cycle_latch = pvs->sound_cycle_latch;

    for (;;)
    {
        unsigned cyc, env_update_time;

        env_update_time = pvs->env_update_time;
        cyc = voice_cycle_latch - pvs->envcyc;

        /* Is it time to adjust envelope? */
        if (cyc < env_update_time)
        {
         /* Should we ever adjust envelope? */
         if (env_update_time == (0 - 1))
          pvs->envcyc = voice_cycle_latch;
         break;
        }

        switch (pvs->envstate)
        {
            case ATTACK:
            if (env_update_time != 0)
            {
                pvs->envcyc += env_update_time;
                envx += ENVX_MAX / 64;      //add 1/64th
            }

            if (env_update_time == 0 || envx >= ENVX_MAX)
            {
                envx = ENVX_MAX;
                if (pvs->sl != ENVX_MAX)
                {
                 pvs->envstate = DECAY;
                 pvs->env_update_time = pvs->dr;
                }
                else
                {
                 pvs->envstate = SUSTAIN;
                 pvs->env_update_time = pvs->sr;
                }
            }
            continue;

            case DECAY:
            if (env_update_time != 0)
            {
                pvs->envcyc += env_update_time;
                envx -= (envx >> 8);    //mult by 1-1/256
            }

            if (env_update_time == 0 || envx <= pvs->sl)
            {
                envx = pvs->sl;
                pvs->envstate = SUSTAIN;
                pvs->env_update_time = pvs->sr;
            }

            continue;

            case SUSTAIN:
            pvs->envcyc += env_update_time;
            envx -= (envx >> 8);    //mult by 1-1/256
            continue;

            case RELEASE:
            //says add 1/256??  That won't release, must be subtract.
            //But how often?  Oh well, who cares, I'll just
            //pick a number. :)
            pvs->envcyc += env_update_time;
            envx -= (ENVX_MAX >> 8);    //sub 1/256th
            if ((envx == 0) || (envx > ENVX_MAX))
            {
                SPC_DSP[DSP_ENDX] |= 1 << voice;
                SNDkeys &= ~(1 << voice);
                envx = 0;
                break;
            }
            continue;

            case INCREASE:
            pvs->envcyc += env_update_time;
            envx += (ENVX_MAX >> 6);    //add 1/64th
            if (envx > ENVX_MAX)
            {
                pvs->envcyc = voice_cycle_latch;
                envx = ENVX_MAX;
                break;
            }
            continue;

            case DECREASE:
            pvs->envcyc += env_update_time;
            envx -= (ENVX_MAX >> 6);    //sub 1/64th
            if (envx > ENVX_MAX)    //underflow
            {
                pvs->envcyc = voice_cycle_latch;
                envx = 0;
                break;
            }
            continue;

            case EXP:
            pvs->envcyc += env_update_time;
            envx -= (envx >> 8);    //mult by 1-1/256
            
            continue;

            case BENT:
            pvs->envcyc += env_update_time;
            if (envx < (ENVX_MAX / 4 * 3))
             envx += ENVX_MAX / 64;     //add 1/64th
            else
             envx += ENVX_MAX / 256;    //add 1/256th
            if (envx > ENVX_MAX)
            {
                pvs->envcyc = voice_cycle_latch;
                envx = ENVX_MAX;
                break;
            }
            continue;

            //case DIRECT:
            //break;
        }
        break;
    }
    pvs->envx = envx;
#ifdef ADSR_OPTIONAL
    if (ADSR_ENABLED)
#endif
#if defined(ADSR) || defined(ADSR_OPTIONAL)
        return envx;
#endif
#ifdef ADSR_OPTIONAL
    else
#endif
#if !defined(ADSR) || defined(ADSR_OPTIONAL)
        return ENVX_MAX;
#endif
}

static void SNDNoteOn(int voices)
{
    int voice, cursamp, adsr1, adsr2, gain;
/*  int lvol,rvol=0,pitch;*/
    struct SAMP_DIR { unsigned short start, loop; } *samp_dir;
    samp_dir = (struct SAMP_DIR *) &SPCRAM[(int)SPC_DSP[DSP_DIR] << 8];
    voices &= SPC_MASK;
    SNDkeys |= voices;
    SPC_DSP[DSP_ENDX] &= ~voices;
    for (voice = 0; voice < 8; voice++)
    {
        if (voices & (1 << voice))
        {
            struct voice_state *pvs;

            pvs = &SNDvoices[voice];
            cursamp = SPC_DSP[(voice << 4) + DSP_VOICE_SRCN];
            pvs->brrptr = samp_dir[cursamp].start;
            pvs->loopptr = &samp_dir[cursamp].loop;
            pvs->bufptr = 0x20000;
            pvs->last2 = pvs->last1 = 0;
            pvs->sound_cycle_latch = TotalCycles;

            adsr1 = SPC_DSP[(voice << 4) + DSP_VOICE_ADSR1];
            if (adsr1 & 0x80)
            {
                //ADSR mode
                adsr2 = SPC_DSP[(voice << 4) + DSP_VOICE_ADSR2];
                pvs->envx = 0;
                pvs->envcyc = TotalCycles;

                pvs->envstate = ATTACK;
                pvs->env_update_time = pvs->ar;
            }
            else
            {
                //GAIN mode
                gain = SPC_DSP[(voice << 4) + DSP_VOICE_GAIN];

                pvs->env_update_time = pvs->gain_update_time;

                if (gain & 0x80)
                {
                    pvs->envx = 0;
                    pvs->envcyc = TotalCycles;

                    pvs->envstate = gain >> 5;

                }
                else
                {
                    pvs->envx = (gain & 0x7F) << 24;
                    pvs->envstate = DIRECT;
                }
            }
        }
    }
}

static void SNDNoteOff(int voices)
{
    int voice;

    voices &= SNDkeys;

    for (voice = 0; voice < 8; voice++)
    {
        if (voices & (1 << voice))
        {
#ifdef ADSR_OPTIONAL
            if(ADSR_ENABLED)
            {
#endif
#if defined(ADSR) || defined(ADSR_OPTIONAL)
            SNDvoices[voice].envstate = RELEASE;
            SNDvoices[voice].env_update_time = 128; /*16 64*/
#endif
#ifdef ADSR_OPTIONAL
            } else {
#endif
#if !defined(ADSR) || defined(ADSR_OPTIONAL)
            SNDkeys &= ~(1 << voice);
#endif
#ifdef ADSR_OPTIONAL
            }
#endif
        }
    }
}

static int sound_enable_mode = 0;

void Reset_Sound_DSP()
{
 int i, samples;

 samples = FREQ / LAG * 2;
 if (sound_enable_mode == 2) samples <<= 1;

 memset(SPC_DSP, 0, 128);
 SPC_MASK = 0xFF;
 SNDkeys = 0;
 sound_output_position = 0;
 block_written[0] = block_written[1] = -1;

 noise = NOISE_FEEDBACK;
 noise_vol = 0;
 noise_freq_factor = 0;
 noise_freq_latch = 0;

 for (i = 0; i < 8; i++)
 {
  SNDvoices[i].envx = 0;
  SNDvoices[i].envcyc = TotalCycles;
  SNDvoices[i].ar = attack_time[0];
  SNDvoices[i].dr = decay_time[0];
  SNDvoices[i].sr = sustain_time[0];
  SNDvoices[i].sl = ENVX_MAX / 8;
  SNDvoices[i].gn = 0;
//SNDvoices[i].envstate = DIRECT;
  SNDvoices[i].env_update_time =
   SNDvoices[i].gain_update_time = (0 - 1);
 }

 if (!sound_enable_mode) return;
 if (!STRsamp) return;

 for (i = 0; i < samples; i++)
 {
  ((output_sample *) STRsamp->data)[i] = OUTPUT_ZERO_BASE;
  ((output_sample *) TMPsamp->data)[i] = OUTPUT_ZERO_BASE;
 }

 voice_set_position(voice_handle, 0);
}

void Remove_Sound()
{
 if (NOISEsamp)
 {
  destroy_sample(NOISEsamp);
  NOISEsamp = 0;
 }
 if (TMPsamp)
 {
  destroy_sample(TMPsamp);
  TMPsamp = 0;
 }
 if (STRsamp)
 {
  if (voice_handle != -1) deallocate_voice(voice_handle);
  destroy_sample(STRsamp);
  STRsamp = 0;
 }

 remove_sound();
 SOUND_ENABLED = sound_enable_mode = 0;
}

int Install_Sound(int stereo)
{
 int samples;
 if (sound_enable_mode) Remove_Sound();

 reserve_voices(1, 0);
 if (install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL))
  return SOUND_ENABLED = sound_enable_mode = 0;

 samples = FREQ / LAG * 2;
 STRsamp = create_sample(16, stereo ? TRUE : FALSE, FREQ, samples);
 TMPsamp = create_sample(16, stereo ? TRUE : FALSE, FREQ, samples);
 NOISEsamp = create_sample(16, FALSE, FREQ, samples);
 if (!STRsamp || !TMPsamp || !NOISEsamp)
 {
  Remove_Sound();
  return SOUND_ENABLED = sound_enable_mode = 0;
 }

 voice_handle = allocate_voice(STRsamp);

 if (voice_handle == -1)
 {
  Remove_Sound();
  return SOUND_ENABLED = sound_enable_mode = 0;
 }

 SOUND_ENABLED = sound_enable_mode = stereo + 1;

 voice_set_playmode(voice_handle, PLAYMODE_LOOP);
 voice_set_volume(voice_handle, 255);
 voice_set_pan(voice_handle, 128);
 Reset_Sound_DSP();
 return SOUND_ENABLED;

}

void update_sound_block(void)
{
 int samples, play_samples, sample_bytes;
 int data_block, write_block;

 samples = FREQ / LAG;
 play_samples = samples;

 if (sound_enable_mode == 2) play_samples <<= 1;

 sample_bytes = sizeof(output_sample[play_samples]);

 /* Write to block that is not playing */
 write_block = voice_get_position(voice_handle) >= samples ? 0 : 1;

 /* But only if we haven't already */
 if (block_written[write_block]) return;
 block_written[write_block] = -1;

 /* Write from block not being written to */
 data_block = sound_output_position >= samples ? 0 : play_samples;

 if (write_block) write_block = play_samples;

 memcpy(((output_sample *) STRsamp->data) + write_block,
  ((output_sample *) TMPsamp->data) + data_block, sample_bytes);
}

void SNDUpdate(void)
{
    unsigned i, from, to, slen;
    output_sample *buf;
    int voice;
    signed char done[8], need_noise = 0;
    struct voice_sample_state
    {
     int outx;
     unsigned step, bufptr;
     short *sbuf;
     signed char lvol, rvol;
     char envx;
    } vss[8];

    if (!SPC_ENABLED || !SOUND_ENABLED) return;

    to = (TotalCycles - sound_cycle_latch) / 64;
    if (!to) return;

    sound_cycle_latch += (TotalCycles - sound_cycle_latch) & ~0x3F;
    from = sound_output_position; to += from;
    slen = TMPsamp->len;
    if (to >= slen) to -= slen;

    // Are we completing a block?
    if ((sound_output_position < (FREQ / LAG) && to >= (FREQ / LAG))
     || (sound_output_position >= (FREQ / LAG) && to < (FREQ / LAG)))
    {
     block_written[0] = block_written[1] = 0;
    }

    sound_output_position = to;

    buf = (output_sample *) TMPsamp->data;

    SNDkeys &= SPC_MASK;
    for (voice = 0; voice < 8; voice++)
    {
        struct voice_sample_state *pvss;
        int lvol, rvol;

        done[voice] = !(SNDkeys & (1 << voice));
        if (done[voice]) continue;

        pvss = &vss[voice];

        lvol =
         ((int) ((signed char) SPC_DSP[(voice << 4) + DSP_VOICE_LVOL])
         * (int) ((signed char) SPC_DSP[DSP_MAIN_LVOL]));

        rvol =
         ((int) ((signed char) SPC_DSP[(voice << 4) + DSP_VOICE_RVOL])
         * (int) ((signed char) SPC_DSP[DSP_MAIN_RVOL]));

        if (SOUND_ENABLED == 2)
        {
            pvss->lvol = lvol >> 7;
            pvss->rvol = rvol >> 7;
        }
        else
        {
            if (lvol < 0) lvol = -lvol;
            if (rvol < 0) rvol = -rvol;

            pvss->lvol = (lvol + rvol) >> 8;
        }

        if (SPC_DSP[DSP_NOV] & (1 << voice))
        {
         pvss->step = (noise_frequencies[SPC_DSP[DSP_FLG] & 0x1F] << 13) / FREQ;
         need_noise = -1;
        }
        else
         pvss->step =
          (((unsigned) *(unsigned short *)&SPC_DSP[(voice << 4) + DSP_VOICE_PITCH_L])
          << 3 << 13) / FREQ;

        pvss->sbuf = SNDvoices[voice].buf;
        pvss->bufptr = SNDvoices[voice].bufptr;
        pvss->outx = 0;

    }

    if (need_noise)
    {
        i = from;
        while (i != to)
        {
//          if (!--noise_freq_latch)
//          {
//              noise_freq_latch = noise_freq_factor;
                /* if output is high, apply xor feedback */
                if (noise < 0)
                {
                    noise <<= 1;
                   noise_vol = noise << 17 >> 17;
                    noise ^= NOISE_FEEDBACK;
                }
                else
                {
                    noise <<= 1;
                }
            if (!--noise_freq_latch)
            {
                noise_freq_latch = noise_freq_factor;
                    noise_vol = rand();//noise << 17 >> 17;
            }

            ((output_sample *) NOISEsamp->data)[i++] = noise_vol;
            if (i >= slen)
             i = 0;
        }
    }

    if (SOUND_ENABLED == 2)
    {
        slen <<= 1; from <<= 1; to <<= 1;
    }

    if (SOUND_ENABLED == 2)
    /* stereo */
    {
        i = from;

        while (i != to)
        {
            int lsample = 0, rsample = 0;
            for (voice = 0; voice < 8; voice++)
            {
                struct voice_sample_state *pvss;
                if (done[voice]) continue;

                pvss = &vss[voice];
                while (pvss->bufptr >= 0x20000)
                {
                    pvss->bufptr -= 0x20000;
                    if (get_chunk(voice))
                    {
                        done[voice] = 1;
                        break;
                    }
                }

                if (done[voice]) continue;

                if (SPC_DSP[DSP_NOV] & (1 << voice))
                    pvss->outx = ((output_sample *) NOISEsamp->data)[i >> 1];
                else
                    pvss->outx = pvss->sbuf[pvss->bufptr >> 13];

                if (pvss->outx != 0)
                {
                    pvss->envx = SNDDoEnv(voice) >> 24;
                    pvss->outx = (pvss->outx * (int) pvss->envx) >> 7;
                }

                SNDvoices[voice].sound_cycle_latch += 64;

                lsample += (pvss->outx * (int) pvss->lvol);
                rsample += (pvss->outx * (int) pvss->rvol);

                pvss->bufptr += pvss->step;
            }

            lsample = OUTPUT_ZERO_BASE + (lsample >> 8);
            if (lsample >= OUTPUT_UPPER_LIMIT)
             buf[i] = OUTPUT_UPPER_LIMIT;
            else if (lsample <= OUTPUT_LOWER_LIMIT)
             buf[i] = OUTPUT_LOWER_LIMIT;
            else
             buf[i] = lsample;
            i++;

            rsample = OUTPUT_ZERO_BASE + (rsample >> 8);
            if (rsample >= OUTPUT_UPPER_LIMIT)
             buf[i] = OUTPUT_UPPER_LIMIT;
            else if (rsample <= OUTPUT_LOWER_LIMIT)
             buf[i] = OUTPUT_LOWER_LIMIT;
            else
             buf[i] = rsample;
            i++;

            if (i >= slen)
             i = 0;
        }
    }
    else
    /* mono */
    {
        i = from;

        while (i != to)
        {
            int sample = 0;
            for (voice = 0; voice < 8; voice++)
            {
                struct voice_sample_state *pvss = &vss[voice];
                if (done[voice]) continue;

                while (pvss->bufptr >= 0x20000)
                {
                    pvss->bufptr -= 0x20000;
                    if (get_chunk(voice))
                    {
                        done[voice] = 1;
                        break;
                    }
                }

                if (done[voice]) continue;

                if (SPC_DSP[DSP_NOV] & (1 << voice))
                    pvss->outx = ((output_sample *) NOISEsamp->data)[i];
                else
                    pvss->outx = pvss->sbuf[pvss->bufptr >> 13];

                if (pvss->outx != 0)
                {
                    pvss->envx = SNDDoEnv(voice) >> 24;
                    pvss->outx = (pvss->outx * (int) pvss->envx) >> 7;
                }

                SNDvoices[voice].sound_cycle_latch += 64;

                sample += (pvss->outx * (int) pvss->lvol);

                pvss->bufptr += pvss->step;
            }

            sample = OUTPUT_ZERO_BASE + (sample >> 8);
            if (sample >= OUTPUT_UPPER_LIMIT)
             buf[i] = OUTPUT_UPPER_LIMIT;
            else if (sample <= OUTPUT_LOWER_LIMIT)
             buf[i] = OUTPUT_LOWER_LIMIT;
            else
             buf[i] = sample;
            i++;

            if (i >= slen)
             i = 0;
        }
    }

    for (voice = 0; voice < 8; voice++)
    {
        if (done[voice])
        {
            SNDkeys &= ~(1 << voice);
            vss[voice].outx = 0;
        }

        SNDvoices[voice].bufptr = vss[voice].bufptr;
        SPC_DSP[(voice << 4) + DSP_VOICE_OUTX] = vss[voice].outx >> 8;
    }
}

void SPC_READ_DSP()
{
    int addr_lo = SPC_DSP_ADDR & 0xF;
    int addr_hi = SPC_DSP_ADDR >> 4;

    /* if we're not reading envx, outx, key on, key off, or endx */
    if (addr_lo != DSP_VOICE_ENVX && addr_lo != DSP_VOICE_OUTX
     && (addr_lo != 0x0C || addr_hi < 4 || addr_hi == 6))
     return;

    SNDUpdate();
    switch(addr_lo)
    {
    case DSP_VOICE_ENVX:
                if (!SOUND_ENABLED) SNDvoices[addr_hi].sound_cycle_latch = TotalCycles;

                if (ENVX_ENABLED && (SNDkeys & (1 << addr_hi)))
                 SPC_DSP[SPC_DSP_ADDR] = SNDDoEnv(addr_hi);
                else
                 SPC_DSP[SPC_DSP_ADDR] = 0;
                break;
    case 0xc:   switch(addr_hi)
                {
                case 0x4:   SPC_DSP[SPC_DSP_ADDR]=SNDkeys;
                            break;
                }
    }
}

void SPC_WRITE_DSP()
{
    int i;
    int addr_lo = SPC_DSP_ADDR & 0xF;
    int addr_hi = SPC_DSP_ADDR >> 4;

    /* if we're not writing to key on, key off, flg, or endx */
    if (addr_lo != 0x0C || addr_hi < 4)
    /* and write would not change data, return */
     if (SPC_DSP[SPC_DSP_ADDR] == SPC_DSP_DATA) return;

    SNDUpdate();
    switch(addr_lo)
    {
    //break just means nothing needs to be done right now
    //Commented cases are unsupported registers
    // These go for individual voices
    case DSP_VOICE_LVOL:    //Volume left
                break;
    case DSP_VOICE_RVOL:    //Volume right
                break;
    case DSP_VOICE_PITCH_L: //Pitch lo
                break;
    case DSP_VOICE_PITCH_H: //Pitch hi
                SPC_DSP[SPC_DSP_ADDR] = SPC_DSP_DATA & 0x3F;
                break;
    case DSP_VOICE_SRCN:    //Source number
                break;
    case DSP_VOICE_ADSR1:   //ADSR1
                if (SPC_DSP_DATA == SPC_DSP[SPC_DSP_ADDR]) break;

                if (!SOUND_ENABLED) SNDvoices[addr_hi].sound_cycle_latch = TotalCycles;
                if (SNDkeys & (1 << addr_hi)) SNDDoEnv(addr_hi);

                SNDvoices[addr_hi].ar = attack_time[SPC_DSP_DATA & 0xF];
                SNDvoices[addr_hi].dr = decay_time[SPC_DSP_DATA >> 4 & 7];

                if (SNDvoices[addr_hi].envstate == ATTACK)
                 SNDvoices[addr_hi].env_update_time =
                  SNDvoices[addr_hi].ar;
                else if (SNDvoices[addr_hi].envstate == DECAY)
                 SNDvoices[addr_hi].env_update_time =
                  SNDvoices[addr_hi].dr;

                if (!(SNDkeys & (1 << addr_hi))) break;

                if (SPC_DSP_DATA & 0x80)
                {
                    //switch to ADSR? not sure what to do
                    if (((SPC_DSP_DATA & 0x80) != (SPC_DSP[SPC_DSP_ADDR] & 0x80)))
                    {
                        SNDvoices[addr_hi].envstate = ATTACK;
                        SNDvoices[addr_hi].env_update_time =
                         SNDvoices[addr_hi].ar;
                    }
                }
                else
                {
                    //switch to a GAIN mode
                    i = SPC_DSP[(addr_hi << 4) + DSP_VOICE_GAIN];

                    SNDvoices[addr_hi].env_update_time =
                     SNDvoices[addr_hi].gain_update_time;

                    if (i & 0x80)
                    {
                        SNDvoices[addr_hi].envstate = i >> 5;

                    }
                    else
                    {
                        SNDvoices[addr_hi].envstate = DIRECT;
                        SNDvoices[addr_hi].envx = (i & 0x7F) << 24;
                    }
                }
                break;
    case DSP_VOICE_ADSR2:   //ADSR2
                if (SPC_DSP_DATA == SPC_DSP[SPC_DSP_ADDR]) break;

                if (!SOUND_ENABLED) SNDvoices[addr_hi].sound_cycle_latch = TotalCycles;
                if (SNDkeys & (1 << addr_hi)) SNDDoEnv(addr_hi);

                SNDvoices[addr_hi].sr = sustain_time[SPC_DSP_DATA & 0x1F];
                SNDvoices[addr_hi].sl = (ENVX_MAX / 8) * ((SPC_DSP_DATA >> 5) + 1);

                if (SNDvoices[addr_hi].envstate == SUSTAIN)
                 SNDvoices[addr_hi].env_update_time =
                  SNDvoices[addr_hi].sr;

                break;
    case DSP_VOICE_GAIN:   //GAIN
                if (SPC_DSP_DATA == SPC_DSP[SPC_DSP_ADDR]) break;

                if (!SOUND_ENABLED) SNDvoices[addr_hi].sound_cycle_latch = TotalCycles;
                if (SNDkeys & (1 << addr_hi)) SNDDoEnv(addr_hi);

                if (SPC_DSP_DATA & 0x80)
                {
                    SNDvoices[addr_hi].gn = SPC_DSP_DATA & 0x1F;

                    if (SNDvoices[addr_hi].envstate == INCREASE
                     || SNDvoices[addr_hi].envstate == DECREASE)
                     SNDvoices[addr_hi].gain_update_time =
                     linear_time[SNDvoices[addr_hi].gn];
                    else if (SNDvoices[addr_hi].envstate == BENT)
                     SNDvoices[addr_hi].gain_update_time =
                     bent_time[SNDvoices[addr_hi].gn];
                    else /*if (SNDvoices[addr_hi].envstate == EXP)*/
                     SNDvoices[addr_hi].gain_update_time =
                     exp_time[SNDvoices[addr_hi].gn];
                }
                else
                {
                    SNDvoices[addr_hi].gain_update_time = (0 - 1);
                }

                /* is voice keyed on? */
                if (!(SNDkeys & (1 << addr_hi))) break;

                /* is gain enabled? */
                if (!(SPC_DSP[(addr_hi << 4) + DSP_VOICE_ADSR1] & 0x80))
                {
                    SNDvoices[addr_hi].env_update_time =
                     SNDvoices[addr_hi].gain_update_time;

                    if (SPC_DSP_DATA & 0x80)
                    {
                        SNDvoices[addr_hi].envstate = SPC_DSP_DATA >> 5;
                    }
                    else
                    {
                        SNDvoices[addr_hi].envstate = DIRECT;
                        SNDvoices[addr_hi].envx = (SPC_DSP_DATA & 0x7F) << 24;
                    }
                }
                break;
    // These are general registers
    case 0xc:
                switch(addr_hi)
                {
                case 0: //Left main volume
                        break;
                case 1: //Right main volume
                        break;
//              case 2: //Left echo volume
//                      break;
//              case 3: //Right echo volume
//                      break;
                case 4: //Key on
                        if (SOUND_ENABLED)
                            SNDNoteOn(SPC_DSP_DATA);
                        break;
                case 5: //Key off
                        if (SOUND_ENABLED)
                            SNDNoteOff(SPC_DSP_DATA);
                        SPC_DSP_DATA = 0;
                        break;
                case 6: //FLG
                        noise_freq_factor = noise_freq_factors[SPC_DSP_DATA & 0x1F];
                        if (noise_freq_latch > noise_freq_factor)
                         noise_freq_latch = noise_freq_factor;

                        break;
                case 7: //ENDX
                        SPC_DSP_DATA = 0;
                        break;
                }
                break;
    case 0xd:
                switch(addr_hi)
                {
//              case 0: //Echo feedback
//                      break;
//              case 1: //Not a valid register
//                      break;
//              case 2: //Modulation
//                      break;
//              case 3: //Noise toggle
//                      break;
//              case 4: //Echo toggle
//                      break;
                case 5: //Source directory address
                        break;
//              case 6: //Echo start address
//                      break;
//              case 7: //Echo delay
//                      break;
                }
//  case 0xf:       //FIR Echo filter
//              break;
    }
    SPC_DSP[SPC_DSP_ADDR] = SPC_DSP_DATA;
}

void sound_pause(void)
{
    if (SOUND_ENABLED) voice_stop(voice_handle);
}

#define VOLUME_MONO 255 //170 - Butcha - prevent from muting mono mode
#define VOLUME_STEREO 255

void sound_resume(void)
{
    switch (SOUND_ENABLED)
    {
     case 2: set_volume(VOLUME_MONO, -1); break;
     case 1: set_volume(VOLUME_STEREO, -1); break;
    }

    if (SOUND_ENABLED) voice_start(voice_handle);
}

BEGIN_DIGI_DRIVER_LIST
 DIGI_DRIVER_SOUNDSCAPE
 DIGI_DRIVER_AUDIODRIVE
 DIGI_DRIVER_WINSOUNDSYS
 DIGI_DRIVER_SB
END_DIGI_DRIVER_LIST

BEGIN_MIDI_DRIVER_LIST
END_MIDI_DRIVER_LIST
