; SPC instruction macros
;  Original code by Savoury SnaX
;  Rewritten/maintained by TRAC

; TCALL: Call through table in pageable 64-bytes of RAM
;%1 = vecnum
%macro TCALL 0-1
ALIGNC
SPC_TCALL_%1:
 GET_PC eax
 inc eax
 PUSH_W
 xor eax,eax
 mov ax,[_SPCRAM+0xFFC0+((15 - (%1)) << 1)]
 cmp eax,0xFFC0
 mov R_NativePC,eax
 mov eax,[SPC_FFC0_Address]
 jnb .possible_rom
 mov eax,_SPCRAM
.possible_rom:
 add R_NativePC,eax
 mov [SPC_Code_Base],eax
 jmp near SPC_BACK
%endmacro

;%1 = label, %2 = flag
%macro SPC_BFC 2
ALIGNC
SPC_%1:
 add R_NativePC,byte 2
 JUMP_FLAG_SPC %2,near SPC_BACK          ; flag set
 movsx eax,byte [-1+R_NativePC]     ; sign extend for addition
 add R_Cycles,byte 2                ; Branch taken
;add dword [_TotalCycles],byte 2    ; (0.32) Butcha - fix 'lost' SPC timer ticks!
 add R_NativePC,eax
 jmp near SPC_BACK
%endmacro

;%1 = label, %2 = flag
%macro SPC_BFS 2
ALIGNC
SPC_%1:
 add R_NativePC,byte 2
 JUMP_NOT_FLAG_SPC %2,near SPC_BACK      ; flag set
 movsx eax,byte [-1+R_NativePC]     ; sign extend for addition
 add R_Cycles,byte 2                ; Branch taken
;add dword [_TotalCycles],byte 2    ; (0.32) Butcha - fix 'lost' SPC timer ticks!
 add R_NativePC,eax
 jmp near SPC_BACK
%endmacro

;%1 = op, %2 = flags
%macro SPC_OP_A_IM 2
 mov al,[1+R_NativePC]
 add R_NativePC,byte 2
 SPC_BASE_%1 al,B_A
 STORE_FLAGS_SPC (%2)
 jmp near SPC_BACK
%endmacro

;%1 = op, %2 = flags
%macro SPC_OP_X_IM 2
 mov al,[1+R_NativePC]
 add R_NativePC,byte 2
 SPC_BASE_%1 al,B_X
 STORE_FLAGS_SPC (%2)
 jmp near SPC_BACK
%endmacro

;%1 = op, %2 = flags
%macro SPC_OP_Y_IM 2
 mov al,[1+R_NativePC]
 add R_NativePC,byte 2
 SPC_BASE_%1 al,B_Y
 STORE_FLAGS_SPC (%2)
 jmp near SPC_BACK
%endmacro

;%1 = op, %2 = flags, %3 = addr
%macro SPC_OP_A 3
 ADDR_%3
 GET_BYTE_SPC
 SPC_BASE_%1 al,B_A
 STORE_FLAGS_SPC (%2)
 jmp near SPC_BACK
%endmacro

;%1 = op, %2 = flags, %3 = addr
%macro SPC_OP_X 3
 ADDR_%3
 GET_BYTE_SPC
 SPC_BASE_%1 al,B_X
 STORE_FLAGS_SPC (%2)
 jmp near SPC_BACK
%endmacro

;%1 = op, %2 = flags, %3 = addr
%macro SPC_OP_Y 3
 ADDR_%3
 GET_BYTE_SPC
 SPC_BASE_%1 al,B_Y
 STORE_FLAGS_SPC (%2)
 jmp near SPC_BACK
%endmacro

;%1 = op, %2 = flags, %3 = addr
%macro SPC_OP_mem 3
 ADDR_%3 cl         ; cl is source byte, bx dest address
 GET_BYTE_SPC       ; al contains byte at dest address
 SPC_BASE_%1 cl,al
 STORE_FLAGS_SPC (%2)
 SET_BYTE_SPC
 jmp near SPC_BACK
%endmacro

;%1 = src, %2 = dest
%macro SPC_BASE_MOV 2
 test byte %1,%1
 mov byte %2,%1
%endmacro

;%1 = src, %2 = dest
%macro SPC_BASE_OR 2
 or %2,%1
%endmacro

;%1 = addr
%macro SPC_OR_A 1
ALIGNC
SPC_OR_A_%1:
 SPC_OP_A OR,SPC_FLAG_NZ,%1
%endmacro

;%1 = addr
%macro SPC_OR_mem 1
ALIGNC
SPC_OR_%1:
 SPC_OP_mem OR,SPC_FLAG_NZ,%1
%endmacro

;%1 = src, %2 = dest
%macro SPC_BASE_AND 2
 and %2,%1
%endmacro

;%1 = addr
%macro SPC_AND_A 1
ALIGNC
SPC_AND_A_%1:
 SPC_OP_A AND,SPC_FLAG_NZ,%1
%endmacro

;%1 = addr
%macro SPC_AND_mem 1
ALIGNC
SPC_AND_%1:
 SPC_OP_mem AND,SPC_FLAG_NZ,%1
%endmacro

;%1 = src, %2 = dest
%macro SPC_BASE_EOR 2
 xor %2,%1
%endmacro

;%1 = addr
%macro SPC_EOR_A 1
ALIGNC
SPC_EOR_A_%1:
 SPC_OP_A EOR,SPC_FLAG_NZ,%1
%endmacro

;%1 = addr
%macro SPC_EOR_mem 1
ALIGNC
SPC_EOR_%1:
 SPC_OP_mem EOR,SPC_FLAG_NZ,%1
%endmacro

;%1 = src, %2 = dest
%macro SPC_BASE_CMP 2
 cmp %2,%1
 cmc
%endmacro

;%1 = reg, %2 = addr
%macro SPC_CMP 2
 ADDR_%2
 GET_BYTE_SPC
 cmp %1,al
 cmc
 STORE_FLAGS_SPC SPC_FLAG_NZC
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_CMP_A 1
ALIGNC
SPC_CMP_A_%1:
 SPC_CMP B_A,%1
%endmacro

;%1 = addr
%macro SPC_CMP_X 1
ALIGNC
SPC_CMP_X_%1:
 SPC_CMP B_X,%1
%endmacro

;%1 = addr
%macro SPC_CMP_Y 1
ALIGNC
SPC_CMP_Y_%1:
 SPC_CMP B_Y,%1
%endmacro

;%1 = addr
%macro SPC_CMP_mem 1
ALIGNC
SPC_CMP_%1:
 ADDR_%1 cl         ; cl is source byte, bx dest address
 GET_BYTE_SPC       ; al contains byte at dest address
 cmp al,cl
 cmc
 STORE_FLAGS_SPC (SPC_FLAG_NZC)
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_ADC_A 1
ALIGNC
SPC_ADC_A_%1:
 ADDR_%1
 CLR_FLAG_SPC SPC_FLAG_V    ; Done cos STORE_FLAGS cant alter V!
 GET_BYTE_SPC
 MAKE_CARRY
 adc B_A,al
 lahf
 jno .no_overflow
 SET_FLAG_SPC SPC_FLAG_V
.no_overflow:
 STORE_FLAGS_SPC_FROM SPC_FLAG_NHZC,ah
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_ADC_mem 1
ALIGNC
SPC_ADC_%1:
 ADDR_%1 cl
 CLR_FLAG_SPC SPC_FLAG_V    ; Done cos STORE_FLAGS cant alter V!
 GET_BYTE_SPC
 MAKE_CARRY
 adc al,cl
 lahf
 jno .no_overflow
 SET_FLAG_SPC SPC_FLAG_V
.no_overflow:
 STORE_FLAGS_SPC_FROM SPC_FLAG_NHZC,ah
 SET_BYTE_SPC
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_SBC_A 1
ALIGNC
SPC_SBC_A_%1:
 ADDR_%1
 CLR_FLAG_SPC SPC_FLAG_V    ; Done cos STORE_FLAGS cant alter V!
 GET_BYTE_SPC
 MAKE_NOT_CARRY
 sbb B_A,al
 cmc
 lahf
 jno .no_overflow
 SET_FLAG_SPC SPC_FLAG_V
.no_overflow:
 STORE_FLAGS_SPC_FROM SPC_FLAG_NHZC,ah
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_SBC_mem 1
ALIGNC
SPC_SBC_%1:
 ADDR_%1 cl
 CLR_FLAG_SPC SPC_FLAG_V    ; Done cos STORE_FLAGS cant alter V!
 GET_BYTE_SPC
 MAKE_NOT_CARRY
 sbb al,cl      ; (0.30) Butcha - switched these, wrong order
 cmc
 lahf
 jno .no_overflow
 SET_FLAG_SPC SPC_FLAG_V
.no_overflow:
 STORE_FLAGS_SPC_FROM SPC_FLAG_NHZC,ah
 SET_BYTE_SPC
 jmp near SPC_BACK
%endmacro

;%1 = reg
%macro SPC_DEC_reg 1
ALIGNC
SPC_DEC_%1:
 inc R_NativePC
 dec byte [__SPC_%1]
 STORE_FLAGS_SPC SPC_FLAG_NZ
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_DEC 1
ALIGNC
SPC_DEC_%1:
 ADDR_%1
 GET_BYTE_SPC
 dec al
 STORE_FLAGS_SPC SPC_FLAG_NZ
 SET_BYTE_SPC
 jmp near SPC_BACK
%endmacro

;%1 = reg
%macro SPC_INC_reg 1
ALIGNC
SPC_INC_%1:
 inc R_NativePC
 inc byte [__SPC_%1]
 STORE_FLAGS_SPC SPC_FLAG_NZ
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_INC 1
ALIGNC
SPC_INC_%1:
 ADDR_%1
 GET_BYTE_SPC
 inc al
 STORE_FLAGS_SPC SPC_FLAG_NZ
 SET_BYTE_SPC
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_ASL 1
ALIGNC
SPC_ASL_%1:
 ADDR_%1
 GET_BYTE_SPC
 add al,al
 STORE_FLAGS_SPC SPC_FLAG_NZC
 SET_BYTE_SPC
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_LSR 1
ALIGNC
SPC_LSR_%1:
 ADDR_%1
 GET_BYTE_SPC
 shr al,byte 1
 STORE_FLAGS_SPC SPC_FLAG_NZC
 SET_BYTE_SPC
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_ROL 1
ALIGNC
SPC_ROL_%1:
 ADDR_%1
 GET_BYTE_SPC
 MAKE_CARRY
 adc al,al
 STORE_FLAGS_SPC SPC_FLAG_NZC
 SET_BYTE_SPC
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_ROR 1
ALIGNC
SPC_ROR_%1:
 ADDR_%1
 GET_BYTE_SPC
 shr R_PSW_B,byte 1     ;MAKE_CARRY
 rcr al,byte 1
 adc R_PSW_B,R_PSW_B    ;STORE_FLAGS_SPC (SPC_FLAG_C)
 test al,al
 STORE_FLAGS_SPC SPC_FLAG_NZ
 SET_BYTE_SPC
 jmp near SPC_BACK
%endmacro

;%1 = dest, %2 = src
%macro SPC_MOV_reg_reg 2
ALIGNC
SPC_MOV_%1__%2:
 inc R_NativePC
 mov al,[__SPC_%2]
 test al,al
 mov [__SPC_%1],al
 STORE_FLAGS_SPC SPC_FLAG_NZ
 jmp near SPC_BACK
%endmacro

;%1 = addr, %2 = reg
%macro SPC_MOV_mem_reg 2
ALIGNC
SPC_MOV_%1__%2:
 ADDR_%1
 mov al,[__SPC_%2]
 SET_BYTE_SPC
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_MOV_mem_A 1
 SPC_MOV_mem_reg %1,A
%endmacro

;%1 = addr
%macro SPC_MOV_mem_X 1
 SPC_MOV_mem_reg %1,X
%endmacro

;%1 = addr
%macro SPC_MOV_mem_Y 1
 SPC_MOV_mem_reg %1,Y
%endmacro

;%1 = reg, %2 = addr
%macro SPC_MOV_reg_mem 2
 ADDR_%2
 GET_BYTE_SPC
 test al,al
 mov %1,al
 STORE_FLAGS_SPC SPC_FLAG_NZ
 jmp near SPC_BACK
%endmacro

;%1 = addr
%macro SPC_MOV_A_mem 1
ALIGNC
SPC_MOV_A_%1:
 SPC_MOV_reg_mem B_A,%1
%endmacro

;%1 = addr
%macro SPC_MOV_X_mem 1
ALIGNC
SPC_MOV_X_%1:
 SPC_MOV_reg_mem B_X,%1
%endmacro

;%1 = addr
%macro SPC_MOV_Y_mem 1
ALIGNC
SPC_MOV_Y_%1:
 SPC_MOV_reg_mem B_Y,%1
%endmacro

;%1 = reg
%macro SPC_PUSH_reg 1
ALIGNC
SPC_PUSH_%1:
 inc R_NativePC
 mov al,[__SPC_%1]
 PUSH_B
 jmp near SPC_BACK
%endmacro

;%1 = reg
%macro SPC_POP_reg 1
ALIGNC
SPC_POP_%1:
 inc R_NativePC
 POP_B
 mov [__SPC_%1],al
 jmp near SPC_BACK
%endmacro
